class ObjectProperties:
    def __init__(self, object_id, extra_data=None, ticket_type=None, quantity=None):
        if extra_data:
            self.extraData = extra_data
        self.objectId = object_id
        if ticket_type:
            self.ticketType = ticket_type
        if quantity:
            self.quantity = quantity



from seatsio.charts.chartsClient import ChartsClient
from seatsio.events.eventsClient import EventsClient
from seatsio.holdtokens.HoldTokenClient import HoldTokensClient
from seatsio.httpClient import HttpClient
from seatsio.reports.usage.usageReports import UsageReports
from seatsio.seasons.seasonsClient import SeasonsClient
from seatsio.subaccounts.subaccountsClient import SubaccountsClient
from seatsio.workspaces.workspacesClient import WorkspacesClient


class Client:
    def __init__(self, region, secret_key, workspace_key=None, max_retries=5):
        base_url = region.url
        self.base_url = base_url
        self.http_client = HttpClient(base_url, secret_key, workspace_key, max_retries)
        self.charts = ChartsClient(self.http_client)
        self.events = EventsClient(self.http_client)
        self.seasons = SeasonsClient(self.http_client, self)
        self.subaccounts = SubaccountsClient(self.http_client)
        self.workspaces = WorkspacesClient(self.http_client)
        self.hold_tokens = HoldTokensClient(self.http_client)
        self.usage_reports = UsageReports(self.http_client)



class SeatsioException(Exception):

    def __init__(self, request, response=None, cause=None):
        if (response is not None) and ("application/json" in response.headers.get("content-type", "")):
            body = response.json()
            self.errors = body["errors"]
            self.requestId = body["requestId"]
            self.cause = cause
            self.message = self.__build_exception_message(request, response)
            super(SeatsioException, self).__init__(self.message)
        else:
            self.errors = None
            self.requestId = None
            self.cause = cause
            self.message = "Error while executing " + request.http_method + " " + request.url
            super(SeatsioException, self).__init__(self.message)

    def __build_exception_message(self, request, response):
        exception_message = request.http_method + " " + request.url + " resulted in a "
        exception_message += str(response.status_code) + " " + str(response.reason) + " response."
        exception_message += " Reason: " + ", ".join(self.__map_errors_to_message(self.errors)) + "."
        exception_message += " Request ID: " + self.requestId
        return exception_message

    def __map_errors_to_message(self, errors):
        return list(map(lambda e: e.get("message"), errors))


class RateLimitExceededException(SeatsioException):

    def __init__(self, request, response=None, cause=None):
        super(RateLimitExceededException, self).__init__(request, response, cause)




class ChartRequest:

    def __init__(self, name=None, venue_type=None, categories=None):
        if name:
            self.name = name
        if venue_type:
            self.venueType = venue_type
        if categories:
            self.categories = categories



from seatsio.pagination.page import Page


class PageFetcher:
    def __init__(self, cls, http_client, url, **kwargs):
        self.cls = cls
        self.url = url
        self.httpClient = http_client
        self.page_size = None
        self.query_params = {}
        self.kwargs = kwargs

    def fetch_after(self, after_id=None):
        if after_id:
            self.set_query_param("start_after_id", after_id)
        return self.__fetch()

    def fetch_before(self, before_id=None):
        if before_id:
            self.set_query_param("end_before_id", before_id)
        return self.__fetch()

    def __fetch(self):
        if self.page_size:
            self.set_query_param("limit", self.page_size)
        response = self.httpClient.url(self.url, self.query_params, **self.kwargs).get()
        return Page.from_response(response, self.cls)

    def set_page_size(self, page_size):
        if page_size is not None:
            self.page_size = page_size
        return self

    def set_query_param(self, key, value):
        if value is not None:
            self.query_params[key] = value
        return self



class CreateMultipleEventsRequest:
    def __init__(self, chart_key, events_properties):
        if chart_key:
            self.chartKey = chart_key
        if events_properties:
            self.events = events_properties



import time

from six.moves.urllib.parse import quote, urlencode

import jsonpickle
import requests

from seatsio.exceptions import SeatsioException, RateLimitExceededException


def handle_error(request, response):
    if response.status_code == 429:
        raise RateLimitExceededException(request, response)
    else:
        raise SeatsioException(request, response)


class HttpClient:
    def __init__(self, base_url, secret_key, workspace_key, max_retries):
        self.base_url = base_url
        self.secret_key = secret_key
        self.workspace_key = workspace_key
        self.max_retries = max_retries

    def url(self, relative_url, query_params=None, **path_params):
        if query_params is None:
            query_params = {}
        return ApiResource(self.max_retries, self.secret_key, self.workspace_key, self.base_url, relative_url, query_params, **path_params)

class ApiResource:
    def __init__(self, max_retries, secret_key, workspace_key, base_url, relative_url, query_params, **path_params):
        self.max_retries = max_retries
        self.url = self.__create_full_url(base_url, relative_url, query_params, **path_params)
        self.secret_key = secret_key
        self.workspace_key = workspace_key

    def __create_full_url(self, base_url, relative_url, query_params, **path_params):
        for key in path_params:
            path_params[key] = quote(str(path_params[key]), safe='')
        full_url = base_url + relative_url.format(**path_params)
        if query_params:
            full_url += "?" + urlencode(query_params, True)
        return full_url

    def get(self):
        return GET(self.max_retries, self.url, self.secret_key, self.workspace_key).execute()

    def get_raw(self):
        return GET(self.max_retries, self.url, self.secret_key, self.workspace_key).execute_raw()

    def get_as(self, cls):
        return cls(self.get())

    def post(self, body=None):
        if body is None:
            return POST(self.max_retries, self.url, self.secret_key, self.workspace_key).execute()
        else:
            return POST(self.max_retries, self.url, self.secret_key, self.workspace_key).body(body).execute()

    def post_empty_and_return(self, cls):
        return cls(self.post().json())

    def delete(self, body=None):
        if body is None:
            return DELETE(self.max_retries, self.url, self.secret_key, self.workspace_key).execute()
        else:
            return DELETE(self.max_retries, self.url, self.secret_key, self.workspace_key).body(body).execute()


class GET:

    def __init__(self, max_retries, url, secret_key, workspace_key):
        self.http_method = "GET"
        self.max_retries = max_retries
        self.url = url
        self.secret_key = secret_key
        self.workspace_key = workspace_key

    def execute(self):
        response = retry(self.try_execute, self.max_retries)
        if response.status_code >= 400:
            handle_error(self, response)
        else:
            return response.json()

    def execute_raw(self):
        response = retry(self.try_execute, self.max_retries)
        if response.status_code >= 400:
            handle_error(self, response)
        else:
            return response.content

    def try_execute(self):
        try:
            return requests.get(self.url, auth=(self.secret_key, ''), headers={'X-Workspace-Key': str(self.workspace_key) if self.workspace_key else None})
        except Exception as cause:
            raise SeatsioException(self, cause=cause)


class POST:

    def __init__(self, max_retries, url, secret_key, workspace_key):
        self.http_method = "POST"
        self.max_retries = max_retries
        self.url = url
        self.secret_key = secret_key
        self.workspace_key = workspace_key
        self.body_object = None

    def body(self, body):
        self.body_object = body
        return self

    def execute(self):
        response = retry(self.try_execute, self.max_retries)
        if response.status_code >= 400:
            handle_error(self, response)
        else:
            return response

    def try_execute(self):
        try:
            json = jsonpickle.encode(self.body_object, unpicklable=False)
            return requests.post(
                url=self.url,
                auth=(self.secret_key, ''),
                headers={'X-Workspace-Key': str(self.workspace_key) if self.workspace_key else None},
                data=json
            )
        except Exception as cause:
            raise SeatsioException(self, cause=cause)


class DELETE:

    def __init__(self, max_retries, url, secret_key, workspace_key):
        self.http_method = "DELETE"
        self.max_retries = max_retries
        self.url = url
        self.secret_key = secret_key
        self.workspace_key = workspace_key
        self.body_object = None

    def body(self, body):
        self.body_object = body
        return self

    def execute(self):
        response = retry(self.try_execute, self.max_retries)
        if response.status_code >= 400:
            handle_error(self, response)
        else:
            return response

    def try_execute(self):
        try:
            json = jsonpickle.encode(self.body_object, unpicklable=False)
            return requests.delete(
                self.url,
                auth=(self.secret_key, ''),
                headers={'X-Workspace-Key': str(self.workspace_key) if self.workspace_key else None},
                data=json
            )
        except Exception as cause:
            raise SeatsioException(self, cause=cause)


def retry(fn, max_retries):
    retry_count = 0
    while True:
        response = fn()
        if response.status_code != 429 or retry_count >= max_retries:
            return response
        else:
            wait_time = (2 ** (retry_count + 2)) / 10.0
            time.sleep(wait_time)
            retry_count += 1
